require "defines"

local easy, medium, hard, master, custom = "easy", "medium", "hard", "master", "custom" 

settings_scheme = medium

custom_scheme = {initial_task_cost = 800, task_cost_mult = 1.2, task_count = 3, max_tech_level = 0.2}

--change seed to get different tasks
map_seed = 96621 --no more than 5 digits!


item_list = {}
disabled_items = {}

--Place here any items, unobtainable by crafting, or with cyclic recipes	
		item_list["iron-ore"] = {name = "iron_ore", loc_name = game.getlocaliseditemname("iron-ore"), cost = 1}
		item_list["copper-ore"] = {name = "copper-ore", loc_name = game.getlocaliseditemname("copper-ore"), cost = 1}
		item_list["coal"] = {name = "coal", loc_name = game.getlocaliseditemname("coal"), cost = 1}
		item_list["stone"] = {name = "stone", loc_name = game.getlocaliseditemname("stone"), cost = 1}
		item_list["crude-oil"] = {name = "crude-oil", loc_name = " ", cost = 5}
		item_list["water"] = {name = "water", loc_name = " ", cost = 1}
		item_list["raw-wood"] = {name = "raw-wood", loc_name = game.getlocaliseditemname("raw-wood"), cost = 10}
		item_list["raw-fish"] = {name = "raw-fish", loc_name = game.getlocaliseditemname("raw-fish"), cost = 100}
		item_list["alien-artifact"] = {name = "alien-artifact", loc_name = game.getlocaliseditemname("alien-artifact"), cost = 500}
		item_list["coin"] = {name = "coin", loc_name = game.getlocaliseditemname("coin"), cost = 1}
		item_list["tank-cannon"] = {name = "tank-cannon", loc_name = game.getlocaliseditemname("tank-cannon"), cost = 1}
		item_list["computer"] = {name = "computer", loc_name = game.getlocaliseditemname("computer"), cost = 1}
		item_list["biter-attractor-ON"] = {name = "biter-attractor-ON", loc_name = game.getlocaliseditemname("biter-attractor-ON"), cost = 1}
		item_list["biter-attractor-OFF"] = {name = "biter-attractor-OFF", loc_name = game.getlocaliseditemname("biter-attractor-OFF"), cost = 1}		
		item_list["pvp-armor"] = {name = "pvp-armor", loc_name = " ", cost = 1}	
		
--Disable any items, that require wood, or other finite resources, small stack size and cheap items (weapons), liquids (only raw), unobtainable items
		disabled_items["crude-oil"] = true
		disabled_items["coin"] = true
		disabled_items["computer"] = true
		disabled_items["water"] = true
		disabled_items["raw-fish"] = true
		disabled_items["raw-wood"] = true
		disabled_items["wood"] = true
		disabled_items["wooden-chest"] = true
		disabled_items["combat-shotgun"] = true
		disabled_items["rocket-launcher"] = true
		disabled_items["shotgun"] = true
		disabled_items["railgun"] = true
		disabled_items["tank-cannon"] = true
		disabled_items["flame-thrower"] = true
		disabled_items["pistol"] = true
		disabled_items["small-electric-pole"] = true
		disabled_items["submachine-gun"] = true
		disabled_items["biter-attractor-ON"] = true
		disabled_items["biter-attractor-OFF"] = true
		disabled_items["pvp-armor"] = true
		disabled_items["pvp-armor-mk2"] = true
		disabled_items["pvp-armor-mk3"] = true
		disabled_items["pvp-armor-mk4"] = true
		disabled_items["pvp-armor-mk5"] = true
		disabled_items["small-plane"] = true
		disabled_items["player-port"] = true
		disabled_items["iron-axe"] = true
		disabled_items["steel-axe"] = true

---------------------------------------SETTTINGS END-----------------------------------------------------------------------
local blue_spawn = {-520,0}
local red_spawn = {520,0}

function init()
  if not glob.deadPlayers then
    glob.deadPlayers = {}
  end
end
game.oninit(init)
game.onload(init)

function Initialize()
  if not glob.initialized then
		Seed()
		game.alwaysday = true
		game.forces.player.researchalltechnologies()
		game.forces.enemy.researchalltechnologies()
		
		if settings_scheme == custom then
			glob.game_props = custom_scheme
		end
		
		if settings_scheme == easy then
			glob.game_props = {initial_task_cost = 800, task_cost_mult = 2, task_count = 3, max_tech_level = 0.2}
		end
		
		if settings_scheme == medium then
			glob.game_props = {initial_task_cost = 2000, task_cost_mult = 2, task_count = 5, max_tech_level = 0.5}
		end
		
		if settings_scheme == hard then
			glob.game_props = {initial_task_cost = 2000, task_cost_mult = 2, task_count = 8, max_tech_level = 0.8}
		end
		
		if settings_scheme == master then
			glob.game_props = {initial_task_cost = 2000, task_cost_mult = 2, task_count = 12, max_tech_level = 1}
		end
		
		glob.game_stats = {red_score = 0, blue_score = 0, current_task = 0, red_count_done = 0, blue_count_done = 0, game_over = false}
		glob.red_task_chest = game.getentitybytag("red_chest")
		glob.blue_task_chest = game.getentitybytag("blue_chest")
		AnalyzeItems()
		GenerateTasks()
		FillStartingChests()
		SetEntityForces()
		PrintAll("Welcome to The Quest map!")
		PrintAll("Current game settings:")
		PrintAll("Map seed: " .. map_seed .. ", Overall difficulty: " .. settings_scheme)
		NextTask()
		glob.initialized = true
	end
end

function GenerateTasks()

	PrintAll("Generating tasks... ")
	glob.tasks = {}
	local tech_step = glob.game_props.max_tech_level/glob.game_props.task_count
	local current_tech = tech_step
	local current_cost = glob.game_props.initial_task_cost
	for	i=1,glob.game_props.task_count do
		local item = GetItem(current_tech - tech_step, current_tech)
		local count = current_cost/item.cost
		if count < 1 then
			count = 1
		end
		count = RoundNumber(count)
		glob.tasks[i] = {item = item, count = count}
		current_tech = current_tech + tech_step
		current_cost = current_cost*glob.game_props.task_cost_mult
	end
	PrintAll("Done! ")
	for key, task in pairs(glob.tasks) do
		PrintAll(task.item.name .. " " .. task.count .. " ")
	end
end

function FillStartingChests()
	local chest_tag = "s_chest_"
	local index = 1
	
	while true do
		local chest = game.getentitybytag("s_chest_".. index)
		if not chest then return end
		chest.insert({name = "steel-axe", count = 5})
		chest.insert({name = "offshore-pump", count = 1})
		chest.insert({name = "boiler", count = 14})
		chest.insert({name = "steam-engine", count = 10})
		chest.insert({name = "pipe", count = 30})
		chest.insert({name = "pipe-to-ground", count = 6})
		
		chest.insert({name = "small-electric-pole", count = 50})
		chest.insert({name = "medium-electric-pole", count = 50})
		
		chest.insert({name = "steel-furnace", count = 20})
		chest.insert({name = "stone-furnace", count = 20})
		chest.insert({name = "burner-mining-drill", count = 15})
		chest.insert({name = "basic-mining-drill", count = 30})
		chest.insert({name = "basic-transport-belt", count = 200})
		chest.insert({name = "basic-splitter", count = 10})
		chest.insert({name = "basic-transport-belt-to-ground", count = 10})
	
		chest.insert({name = "burner-inserter", count = 50})
		chest.insert({name = "basic-inserter", count = 50})
		chest.insert({name = "long-handed-inserter", count = 20})
		chest.insert({name = "fast-inserter", count = 10})
		chest.insert({name = "assembling-machine-1", count = 10})
		chest.insert({name = "assembling-machine-2", count = 10})
		
		chest.insert({name = "raw-wood", count = 20})
		chest.insert({name = "coal", count = 100})
		chest.insert({name = "iron-plate", count = 100})
		chest.insert({name = "copper-plate", count = 100})
		chest.insert({name = "stone", count = 100})
		chest.insert({name = "iron-gear-wheel", count = 100})
		chest.insert({name = "electronic-circuit", count = 100})
		index = index + 1
	end
end

function SetEntityForces()
	local player_entities = game.findentities({{-700,-400},{20,400}}) --left map side
	local enemy_entities = game.findentities({{20,-400},{700,400}}) --right map side
	for	i, entity in pairs(player_entities) do
		entity.force = game.forces.player
	end
	for	i, entity in pairs(enemy_entities) do
		entity.force = game.forces.enemy
	end
end

function RoundNumber(number)
	local order = math.floor(math.log10(number))
	local base = math.pow(10, order)
	local multiplier = math.floor(number/base)
	return (base*multiplier)
end

function Seed()
	for i = 1, map_seed do
		math.random()
	end
end

function NextTask()
if not glob.game_stats.game_over then
	glob.game_stats.current_task = glob.game_stats.current_task + 1
	if glob.game_stats.current_task > glob.game_props.task_count then
	
		if glob.game_stats.red_score > glob.game_stats.blue_score then
			PrintAll("Game is over! Red team WINS!")
		end
		if glob.game_stats.red_score < glob.game_stats.blue_score then
			PrintAll("Game is over! Blue team WINS!")
		end
		if glob.game_stats.red_score == glob.game_stats.blue_score then
			PrintAll("Game is over! It's a tie!")
		end
		PrintAll("Final score B:" ..  glob.game_stats.blue_score .. ", R:" ..  glob.game_stats.red_score)
		glob.game_stats.game_over = true
	else
		PrintAll("Next task!")
		local tasks_left = glob.game_props.task_count - glob.game_stats.current_task + 1
		PrintAll("Tasks to go:" .. tasks_left)
	end
	UpdateGUI()	
end
end

function UpdateGUI()

	local Blue = {b=1, g = 0.5}
	local Red = {r=1, b = 0.2, g = 0.2}
   	for i, player in ipairs(game.players) do
		if game.players[i].gui.top["red_score"] ~= nil then
			game.players[i].gui.top["blue_score"].destroy()
			game.players[i].gui.top["score"].destroy()
			game.players[i].gui.top["colon"].destroy()
			game.players[i].gui.top["red_score"].destroy()
			game.players[i].gui.top["current_task"].destroy()
			game.players[i].gui.top["current_task_text"].destroy()
		end
		game.players[i].gui.top.add{type="label", name="score", caption= "Score ", style="caption_label_style"}
		game.players[i].gui.top.add{type="label", name="blue_score", caption= glob.game_stats.blue_score, style="caption_label_style"}.style.fontcolor = Blue
		game.players[i].gui.top.add{type="label", name="colon", caption= ":", style="caption_label_style"}
		game.players[i].gui.top.add{type="label", name="red_score", caption= glob.game_stats.red_score, style="caption_label_style"}.style.fontcolor = Red
		if not glob.game_stats.game_over then
			game.players[i].gui.top.add{type="label", name="current_task_text", caption= "| Current task: ", style="caption_label_style"}
			game.players[i].gui.top.add{type="label", name="current_task", caption= {"", glob.tasks[glob.game_stats.current_task].count, "  ", {glob.tasks[glob.game_stats.current_task].item.loc_name[1]}, "|"}, style="caption_label_style"}
		end
	end
end

function GetItem(min_tech, max_tech)
	local tech = min_tech + math.random()*(max_tech - min_tech)
	local target_cost = math.pow(glob.max_cost, tech) 
	local closest_item = item_list["iron-ore"]
	local min_distance = Distance(item_list["iron-ore"].cost, target_cost)
	for	i, item in pairs(item_list) do
		if Distance(item.cost, target_cost) < min_distance then
			if not disabled_items[item.name] then
				min_distance = Distance(item.cost, target_cost) 
				closest_item = item
			end
		end
	end
	return closest_item 
end

function Distance(x1, x2)
return math.abs(x1 - x2)
end

function AnalyzeItems()
	
	PrintAll("Obtaining items info... ")
	glob.max_cost = 1
	for key, item in pairs(game.itemprototypes) do
		GetItemCost(item.name)	
	end
	--GetItemComplexity("power-armor-mk2")
	PrintAll("Done!")
	
	--for key, item in pairs(item_list) do
	--	PrintAll(key .. " " .. " " .. item.cost)	
	--end
end

function GetFluidCost(name)
	if not item_list[name] then
		local new_fluid = {name = " ", cost = 0}
		local recipe, result_count = GetRecipeFor(name)
		local ing_complexity 
		for	i, ing in pairs(recipe.ingredients) do
			if ing.type == 1 then
				ingredient_cost = GetFluidCost(ing.name)
			else	
				ingredient_cost = GetItemCost(ing.name)
			end
				new_fluid.cost = ingredient_cost.cost*ing.amount + new_fluid.cost
			end
		new_fluid.cost = new_fluid.cost/result_count
		return new_fluid
	end
	return item_list[name]
end

function GetItemCost(name)
		if not item_list[name] then
			local new_item = {name = name, loc_name = game.getlocaliseditemname(name), cost = 0}
			local recipe, result_count = GetRecipeFor(name)
			local ingredient_cost 
			for	i, ing in pairs(recipe.ingredients) do
				if ing.type == 1 then
					ingredient_cost = GetFluidCost(ing.name)
				else	
					ingredient_cost = GetItemCost(ing.name)
				end
				new_item.cost = ingredient_cost.cost*ing.amount + new_item.cost
			end
			new_item.cost = new_item.cost/result_count
			item_list[name] = new_item
		end
		if item_list[name].cost > glob.max_cost then
			glob.max_cost = item_list[name].cost
		end
		return item_list[name]
end

function GetRecipeFor(name)
	local recipe
	local result_count
	local found
	if not game.forces.player.recipes[name] then
		for i, rec in pairs(game.forces.player.recipes) do
			result_count, found = FindInProducts(rec, name)
			if found then 
				recipe = rec
				break
			end
		end
	else	
		recipe = game.forces.player.recipes[name]
		result_count, found = FindInProducts(recipe, name)
	end
	if not recipe or not result_count then
		PrintAll("Error, while trying to find recipe for " .. name)
	end
	return recipe, result_count	
end

function FindInProducts(recipe, name)
	local found = false
	local result_count
	for j, result in pairs(recipe.products) do
		if result.name == name then
			result_count = result.amount
			found = true
			break
		end
	end
	return result_count, found
end

function PrintAll(text)
	for i, player in pairs(game.players) do
		player.print(text)
	end
end

function Spawn(player_index, player)
	if player_index % 2 == 0 then
   		local position = game.findnoncollidingposition("player", red_spawn , 50, 1)
			if position ~= nil then player.teleport(position) end
	else
   		local position = game.findnoncollidingposition("player", blue_spawn , 50, 1)
			if position ~= nil then player.teleport(position) end
	end
   	glob.deadPlayers[player_index] = nil
end


game.onevent(defines.events.onplayerrotatedentity, function(event)
if not glob.game_stats.game_over then
	if event.entity.equals(glob.red_task_chest) then
		local count = event.entity.getitemcount(glob.tasks[glob.game_stats.current_task].item.name)
		glob.game_stats.red_count_done = glob.game_stats.red_count_done + count
		event.entity.getinventory(1).clear()
		if (glob.game_stats.red_count_done >= glob.tasks[glob.game_stats.current_task].count) then
			PrintAll("Red team scores!")
			glob.game_stats.red_count_done = 0
			glob.game_stats.red_score = glob.game_stats.red_score + 1	
			NextTask()
		end	
	end
	if event.entity.equals(glob.blue_task_chest) then
		local count = event.entity.getitemcount(glob.tasks[glob.game_stats.current_task].item.name)
		glob.game_stats.blue_count_done = glob.game_stats.blue_count_done + count
		event.entity.getinventory(1).clear()
		if (glob.game_stats.blue_count_done >= glob.tasks[glob.game_stats.current_task].count) then
			PrintAll("Blue team scores!")
			glob.game_stats.blue_count_done = 0
			glob.game_stats.blue_score = glob.game_stats.blue_score + 1
			NextTask()
		end
	end
end	
end)

game.onevent(defines.events.onentitydied, function(event)
  if event.entity.type == "player" then
    for index, player in ipairs(game.players) do
      if not player.character and not glob.deadPlayers[index] then
        glob.deadPlayers[index] = true
        break
      end
    end
  end
end)

game.onevent(defines.events.ontick, function(event)
	for playerIndex, value in pairs(glob.deadPlayers) do
	local player = game.players[playerIndex]
   		if player.character then
			Spawn(playerIndex, player)
   		end
	end
end)


game.onevent(defines.events.onplayercreated, function(event)
	Initialize()
	UpdateGUI()
	local player = game.players[event.playerindex]
   		if player.character then
			Spawn(event.playerindex, player)
   		end
end)